#include <time.h>
#include <extdll.h>
#include <sdk_util.h>

#include <dllapi.h>
#include <meta_api.h>

#include <pm_defs.h>

#include "vers_plugin.h"
#include "mtrand.h"


float player_timer[33];


static MTRandClass RandGen = MTRandClass(1234);


static const char *GetModel(edict_t *pEntity) 
{
	if(!FNullEnt(pEntity))
		return( g_engfuncs.pfnInfoKeyValue( (*g_engfuncs.pfnGetInfoKeyBuffer)( pEntity ), "model" ) );
	
	static char retval[1]; retval[0]=0; return retval;
}


//Sends message to player
static int message_TextMsg = 0;
static void SendTextMsg(edict_t *pEntity, int msg_dest, const char* msg_name)
{
	if (FNullEnt(pEntity)) return;
	
	if (message_TextMsg == 0) message_TextMsg = GET_USER_MSG_ID(PLID,"TextMsg",NULL);
	
	if (message_TextMsg == 0 || pEntity->v.flags & FL_FAKECLIENT) return;
	
	MESSAGE_BEGIN( MSG_ONE, message_TextMsg, NULL, pEntity );
		WRITE_BYTE( msg_dest );
		WRITE_STRING( msg_name );
	MESSAGE_END();
}


static void ClientCommand(edict_t *pEntity)
{
	try {
		int pid = ENTINDEX(pEntity);
		if(pid <= 0 || pid > 32)
		{
			RETURN_META(MRES_IGNORED);
		}
		
		//Check if dead or in readyroom
		if((pEntity->v.deadflag != DEAD_NO) || (pEntity->v.health <= 0) || (pEntity->v.flags & FL_NOTARGET) || pEntity->v.team == 0)
		{
			RETURN_META(MRES_IGNORED);
		}

		if (stricmp(CMD_ARGV(0),"say")!=0 && stricmp(CMD_ARGV(0),"say_team")!=0) 
		{
			RETURN_META(MRES_IGNORED);
		}
	
		if(strnicmp(CMD_ARGV(1),"/unstuck",8)!=0 && strnicmp(CMD_ARGV(1),"/stuck",6)!=0)
		{
			RETURN_META(MRES_IGNORED);
		}
		
		int usehull = 0;
	
		//Check which hull should be used
		const char *model = GetModel(pEntity);
		if(!strnicmp(model,"alien",5))
		{
			int model_5 = model[5];
		
			if((model_5=='1' || model_5=='2' || model_5=='3') && model[6]==0)
			{
				//Small aliens (always duck hull)
				usehull = (int)head_hull;
			}
			else if(model_5=='4' && model[6]==0)
			{
				//Check if ducking
				usehull = ( pEntity->v.flags & FL_DUCKING ) ? (int)head_hull : (int)human_hull;
			}
			else if(model_5=='5' && model[6]==0)
			{
				//Check if onos ducking
				usehull = ( pEntity->v.flags & FL_DUCKING ) ? (int)human_hull : (int)large_hull;
			}
			else
			{
				RETURN_META(MRES_IGNORED);
			}
		}
		else if(!stricmp(model,"soldier") || !stricmp(model,"heavy"))
		{
			//Check if ducking
			usehull = ( pEntity->v.flags & FL_DUCKING ) ? (int)head_hull : (int)human_hull;
		}
		else
		{
			//skip commander and gestate
			RETURN_META(MRES_IGNORED);
		}

		//Check timer
		float difference = (player_timer[pid] + CVAR_GET_FLOAT("unstuck_timer")) - gpGlobals->time;
		if( difference > 0 )
		{
			char string[256];
			
			sprintf(string, "[%s] Cannot unstuck yet! Wait %.2f sec and try again!\n", VLOGTAG, difference);
			
			SendTextMsg(pEntity, HUD_PRINTTALK, string);
			
			RETURN_META(MRES_IGNORED);
		}

		TraceResult tr;
		const Vector &origin = pEntity->v.origin;
	
		float height;
		float width;
		
		switch(usehull)
		{
			case (int)human_hull:
				height=72.0f;
				width=32.0f;
			break;
			
			case (int)large_hull:
				height=108.0f;
				width=64.0f;
			break;
			
			case (int)head_hull:
			default:
				height=36.0f;
				width=32.0f;
			break;
		}
	
		height *= 1.5f;
		width *= 1.5f;
	
		float half_height = height * 0.5f;
		float half_width = width * 0.5f;
	
		Vector new_origin;
		
		int type = 0;

		for(int i = 0; i < 6*128; ++i)
		{
			float x,y,z;
		
			switch(type++)
			{
				case 0: 
			   		x = half_width + RandGen.RandomFloat(0,half_width);
			   		y = RandGen.RandomFloat(-width,width);
			   		z = RandGen.RandomFloat(-height,height);
				break;
			
				case 1: 
			   		x = -half_width - RandGen.RandomFloat(0,half_width);
			   		y = RandGen.RandomFloat(-width,width);
			   		z = RandGen.RandomFloat(-height,height);
				break;
			
				case 2: 
			   		x = RandGen.RandomFloat(-width,width);
			   		y = half_width + RandGen.RandomFloat(0,half_width);
			   		z = RandGen.RandomFloat(-height,height);
				break;
			
				case 3: 
					x = RandGen.RandomFloat(-width,width);
			   		y = -half_width - RandGen.RandomFloat(0,half_width);
			   		z = RandGen.RandomFloat(-height,height);
				break;
			
				case 4: 
			   		x = RandGen.RandomFloat(-width,width);
			   		y = RandGen.RandomFloat(-width,width);
			   		z = half_height + RandGen.RandomFloat(0,half_height);
				break;
			
				default:
			   		x = RandGen.RandomFloat(-width,width);
			   		y = RandGen.RandomFloat(-width,width);
			   		z = -half_height - RandGen.RandomFloat(0,half_height);
			   
				   	type = 0;
				break;
			}
		
			new_origin = origin + Vector( x,y,z );
		
			TRACE_LINE( origin, new_origin, FALSE, NULL, &tr );
		
			if( tr.flFraction >= 0.998f )
			{
				TRACE_HULL( new_origin, new_origin, FALSE, usehull, NULL, &tr );
			
				if(!tr.fStartSolid)
				{
					pEntity->v.origin = new_origin;
					player_timer[pid] = gpGlobals->time;
					break;
				}
			}
		}
	
	} 
	catch(...) 
	{ 
		RETURN_META(MRES_IGNORED);
	}
	
	RETURN_META(MRES_HANDLED);
}


static void ServerDeactivate(void)
{
	try {
		for(unsigned i = 0; i < (sizeof(player_timer)/sizeof(player_timer[0])); ++i)
			player_timer[i] = 0.0f;
	} 
	catch(...) 
	{ 
		RETURN_META(MRES_IGNORED);
	}
		
	RETURN_META(MRES_HANDLED);
}


extern "C" int Get_Entity_API2( DLL_FUNCTIONS *pFunctionTable, int *interfaceVersion )
{
	try {
		if( !pFunctionTable )
		{
			return( FALSE );
		}
		else if( *interfaceVersion != INTERFACE_VERSION )
		{
			*interfaceVersion = INTERFACE_VERSION;
			return( FALSE );
		}
	
		memset(pFunctionTable,0,sizeof(DLL_FUNCTIONS));

		pFunctionTable->pfnClientCommand    = ClientCommand;
		pFunctionTable->pfnServerDeactivate = ServerDeactivate;
	}
	catch(...)
	{
		return( FALSE );
	}
	
	return( TRUE );
}
